import json
import numbers
from watson_transformer.contrib.response_base import ResponseBase
from pyspark.sql.types import StringType, FloatType, StructType, StructField, Row
from pyspark.sql import functions as F


"""
"
" default STT output interpreter which assume one alternative transcript and no speaker detection
"
"""

class DefaultNLUParser(ResponseBase):

    def __init__(self, keywords_limit, concepts_limit):
        """
        @param::keywords_limit: the max number of keywords extracted
        @param::concepts_limit: the max number of concepts extracted
        @return: none
        """
        super(DefaultNLUParser, self).__init__()
        if not isinstance(keywords_limit, numbers.Number):
            raise ValueError('> DefaultNLUParser: keywords_limit must be numeric.')
        if not isinstance(concepts_limit, numbers.Number):
            raise ValueError('> DefaultNLUParser: concepts_limit must be numeric.')
        if keywords_limit <= 0:
            raise ValueError('> DefaultNLUParser: keywords_limit must be greater than 0.')
        if concepts_limit <= 0:
            raise ValueError('> DefaultNLUParser: concepts_limit must be greater than 0.')
        self.keywords_limit = keywords_limit
        self.concepts_limit = concepts_limit

    """
    "
    " default NLU output formatter which parse keywords, concepts, sentiment and emotion
    "
    """
    def __call__(self, json_dumps):
        """
        @param::output: the output json object from STT
        @return:the transcript join by period in string format
        """
        data = {}
        valid_json_dumps = True
        try:
            json_data = json.loads(json_dumps)
        except:
            valid_json_dumps = False

        if valid_json_dumps and json_data:
            # extract keyword data
            if "keywords" in json_data:
                for i in range(self.keywords_limit):
                    if i < len(json_data["keywords"]):
                        kw = json_data["keywords"][i]
                        data['keyword_%d'%(i)] = kw['text']
                        data['keyword_%d_score'%(i)] = kw['relevance']
                    else: # when less keywords extracted than limit
                        data['keyword_%d'%(i)] = None
                        data['keyword_%d_score'%(i)] = None
            else:
                for i in range(self.keywords_limit):
                    data['keyword_%d'%(i)] = None
                    data['keyword_%d_score'%(i)] = None

            # extract concept
            if "concepts" in json_data:
                for i in range(self.concepts_limit):
                    if i < len(json_data["concepts"]):
                        concept = json_data["concepts"][i]
                        data['concept_%d'%(i)] = concept['text']
                        data['concept_%d_score'%(i)] = concept['relevance']
                    else: # when less concept extracted than limit
                        data['concept_%d'%(i)] = None
                        data['concept_%d_score'%(i)] = None
            else:
                for i in range(self.concepts_limit):
                    data['concept_%d'%(i)] = None
                    data['concept_%d_score'%(i)] = None

            # extract sentiment
            if "sentiment" in json_data:
                data["sentiment_score"] = json_data["sentiment"]["document"]["score"]
                data["sentiment_label"] = json_data["sentiment"]["document"]["label"]
            else:
                data["sentiment_score"] = None
                data["sentiment_label"] = None

            # extract "emotion"
            if "emotion" in json_data:
                data["sadness_score"] = json_data["emotion"]["document"]["emotion"]["sadness"]
                data["joy_score"] = json_data["emotion"]["document"]["emotion"]["joy"]
                data["fear_score"] = json_data["emotion"]["document"]["emotion"]["fear"]
                data["disgust_score"] = json_data["emotion"]["document"]["emotion"]["disgust"]
                data["anger_score"] = json_data["emotion"]["document"]["emotion"]["anger"]
            else:
                data["sadness_score"] = None
                data["joy_score"] = None
                data["fear_score"] = None
                data["disgust_score"] = None
                data["anger_score"] = None
        
        # case 1: json dumps is invalid
        # case 2: json dumps is valid but json data is invalid for the parser
        if not data:
            # extract keyword data
            for i in range(self.keywords_limit):
                data['keyword_%d'%(i)] = None
                data['keyword_%d_score'%(i)] = None
            # extract concept
            for i in range(self.concepts_limit):
                data['concept_%d'%(i)] = None
                data['concept_%d_score'%(i)] = None
            # extract sentiment
            data["sentiment_score"] = None
            data["sentiment_label"] = None
            # extract "emotion"
            data["sadness_score"] = None
            data["joy_score"] = None
            data["fear_score"] = None
            data["disgust_score"] = None
            data["anger_score"] = None
            
        return Row(**data)

    """
    "
    " return the default NLU return type
    "
    """
    def get_return_type(self):
        """
        @param::num_keywords: the number of keywords extracted by NLU
        @param::num_concpets: the number of concepts extracted by NLU
        @return: the defined return type and Pandas UDF data type
        """
        fields = []
        # populate keyword fields
        for i in range(self.keywords_limit):
            fields.append(StructField("keyword_%d"%(i), StringType(), True))
            fields.append(StructField("keyword_%d_score"%(i), FloatType(), True))

        # populate concpet fields
        for i in range(self.concepts_limit):
            fields.append(StructField("concept_%d"%(i), StringType(), True))
            fields.append(StructField("concept_%d_score"%(i), FloatType(), True))

        # populate other fields
        fields.extend([StructField("sentiment_score", FloatType(), True),
                    StructField("sentiment_label", StringType(), True),
                    StructField("sadness_score", FloatType(), True),
                    StructField("joy_score", FloatType(), True),
                    StructField("fear_score", FloatType(), True),
                    StructField("disgust_score", FloatType(), True),
                    StructField("anger_score", FloatType(), True)])

        return StructType(fields)



"""
"
" pyspark transformer consume IBM Watson service
" @ref: https://stackoverflow.com/questions/41399399/serialize-a-custom-transformer-using-python-to-be-used-within-a-pyspark-ml-pipel/52467470#52467470
" @ref: https://stackoverflow.com/questions/32331848/create-a-custom-transformer-in-pyspark-ml
"
"""
import os
import numbers
import pandas as pd
from concurrent.futures import ThreadPoolExecutor
from pyspark import keyword_only
from pyspark.sql import functions as F
from pyspark.sql.types import StringType
from pyspark.sql import DataFrame
from pyspark.ml.pipeline import Transformer
from pyspark.ml.param.shared import HasInputCol, HasOutputCol, Param
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable


class WatsonServiceTransformer(Transformer, 
                               HasInputCol, 
                               HasOutputCol,
                               DefaultParamsReadable,
                               DefaultParamsWritable):
    
    """
    "
    " set init transformer and set parameters
    "
    """
    @keyword_only
    def __init__(self, 
                 inputCol=None, 
                 outputCol=None, 
                 vectorization=False,
                 max_workers=5,
                 service=None):
        """
        @param::inputCol: the input column name contains sound file name
        @param::outputCol: the output column name
        @param::vectorizaton: flag indicate whether enable vectorization
        @param::max_workders: the max number of workers for each task
        @param::service: the IBM service object
        @return: none
        """

        super(WatsonServiceTransformer, self).__init__()
        self.service = Param(self, "service", None)
        self._setDefault(service=None)
        self.vectorization = Param(self, "vectorization", False)
        self._setDefault(vectorization=False)
        self.max_workers = Param(self, "max_workers", 5)
        self._setDefault(max_workers=5)
        kwargs = self._input_kwargs
        self._set(**kwargs)

        # make sure parameter: token, endpoint set properly.
        if not callable(service):
            raise ValueError('> The service instance provided must be callable object.')
        if not isinstance(max_workers, numbers.Number) or max_workers <= 0:
            raise ValueError('> The number of maximum workers must greater than 0.')
        if not inputCol or not inputCol.strip():
            raise ValueError('> The input column name is required.')
        if not outputCol or not outputCol.strip():
            raise ValueError('> The output column name is required.')
    
    """
    "
    " set parameters, called from self._set(), inherited method.
    "
    """
    @keyword_only
    def setParams(self):
        """
        @param:: None
        @return: none
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    """
    "
    " set whether or not enable vectorized udf
    "
    """
    def setVectorization(self, value):
        """
        @param::value: boolean value indcating enable vectorized udf 
        @return: None
        """
        return self._set(vectorization=value)
    
    """
    "
    " get enable state of vectorized udf
    "
    """
    def getVectorization(self):
        """
        @param::None
        @return: vectorization flag
        """
        return self.getOrDefault(self.vectorization)

    """
    "
    " set the maximum numbers of workder in each task
    "
    """
    def setMax_workers(self, value):
        """
        @param::value: unsigned int indicate the max number of workers
        @return: None
        """
        return self._set(max_workers=value)
    
    """
    "
    " get the max number of workers in each task
    "
    """
    def getMax_workers(self):
        """
        @param::None
        @return: the configured max workers
        """
        return self.getOrDefault(self.max_workers)
    
    """
    "
    " set the API service object
    "
    """
    def setService(self, value):
        """
        @param::value: the IBM service API object
        @return: None
        """
        return self._set(service=value)
    
    """
    "
    " get the API service object
    "
    """
    def getService(self):
        """
        @param:None
        @return: the configured service object
        """
        return self.getOrDefault(self.service)

    """
    "
    " set input columns name
    "
    """
    def setInputCol(self, value):
        """
        @param::value: the input columns name contains the sound file name
        @return: None
        """
        return self._set(inputCol=value)

    """
    "
    " set output columns name
    "
    """
    def setOutputCol(self, value):
        """
        @param::value: the output column name contains the output result
        @return: None
        """
        return self._set(outputCol=value)

    
    """
    "
    " perform the transform using provided IBM service api
    "
    " exploit arrow and vectorized udf
    " ref: https://spark.apache.org/docs/latest/sql-pyspark-pandas-with-arrow.html 
    "
    " to prevent udf from being called multiple times, use asNondeterministic()
    " issue: https://github.com/apache/spark/pull/19929/files/cc309b0ce2496365afd8c602c282e3d84aeed940
    " ref:https://stackoverflow.com/questions/58696198/spark-udf-executed-many-times
    "
    """
    def _transform(self, df:DataFrame) -> DataFrame:
        """
        @param::df: the pyspark dataframe
        @return: the transformed dataframe
        """
        # get the new service instance
        service = self.getService()
        enable_vectorization = self.getVectorization()
        max_workers = max(self.getMax_workers(), 1)
        return_type = service.get_return_type()

        # define the (Vectorized) UDF
        if enable_vectorization:
            # vectorized udf
            @F.pandas_udf(return_type, F.PandasUDFType.SCALAR)
            def vectorized_udf(input_data):
                os.environ['ARROW_PRE_0_15_IPC_FORMAT']='1'
                results = []
                with ThreadPoolExecutor(max_workers=max_workers) as executor:
                    results = executor.map(lambda data:service.get_new_client()(data), input_data)
                return pd.Series(results)
        else:
            # regular udf
            default_udf = F.udf(lambda data:service.get_new_client()(data), return_type)
        udf = vectorized_udf if enable_vectorization else default_udf
        udf = udf.asNondeterministic() # prevent udf from being called mutliple times
        df = df.withColumn(self.getOutputCol(), udf(F.col(self.getInputCol())))
        return df



"""
"
" IBM NLU Service Executable Class
" docu: https://cloud.ibm.com/apidocs/natural-language-understanding?code=python#analyze-text
"
"""

import json
from pyspark import keyword_only
from ibm_watson import NaturalLanguageUnderstandingV1
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_cloud_sdk_core import ApiException
from ibm_watson.natural_language_understanding_v1 import Features, KeywordsOptions, ConceptsOptions, SentimentOptions, EmotionOptions
from pyspark.sql.types import StringType
from watson_transformer.service.service_base import ServiceBase

class NLU(ServiceBase):
    
    @keyword_only
    def __init__(self, token, endpoint, strict_mode=True, **params):
        """
        @param::token: the IBM NLU API access token
        @param::endpoint: the endpoint url for the NLU API
        @param::params: the kv params passing to underlying NaturalLanguageUnderstandingV1 constructor
        @return: the output parsed by parser object
        """
        super(NLU, self).__init__(strict_mode)
        self.token = token
        self.endpoint = endpoint
        self.params = params
        
    def __call__(self, text):
        """
        @param::text: the text to perform NLU
        @return: the output formatted by formatter object
        """

        if text:
            # init nlu client
            authenticator = IAMAuthenticator(self.token)
            nlu = NaturalLanguageUnderstandingV1( version='2019-07-12',authenticator=authenticator)
            nlu.set_service_url(self.endpoint)
            
            try:
                response = nlu.analyze(text = text, **self.params).get_result()
            except ApiException:
                response = None # better to log such execeptions separately
            except Exception:
                if self.strict_mode:
                    raise RuntimeError("*** runtime error caused by input: '%s'"%(text))
                else:
                    response = None
            
            return json.dumps(response) if response else None
        else:
            return None
    
    def get_return_type(self):
        return StringType()

    def get_new_client(self):
        return NLU(token = self.token, 
                   endpoint = self.endpoint,
                   strict_mode = self.strict_mode,
                   **self.params)


"""
"
" test watson service transformer class
"
"""
import pytest
from unittest import mock
from watson_transformer.json_transformer import JSONTransformer



@pytest.fixture(scope='function')
def mock_service(request):
    # mock the service
    service = mock.MagicMock(return_value='foo response')
    return lambda : service


class TestJSONTransformer():

    def test_init_input_col_with__valid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        column_name = "input_column_name"
        # act
        for column_name in ["foo", "foo bar", "foo_bar", "   _"]:
            transformer = JSONTransformer(inputCol=column_name, 
                                          outputCol='output_column',
                                          removeInputCol=True,
                                          parser=mocked_service)
            # assert
            assert transformer.getInputCol() == column_name

    def test_init_input_col_with_invalid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        column_name = None
        # act
        for column_name in [None, "", " ", "    "]:
            with pytest.raises(ValueError) as exception:
                transformer = JSONTransformer(inputCol=column_name, 
                                          outputCol='output_column',
                                          removeInputCol=True,
                                          parser=mocked_service)
            # assert
            assert "input column name" in str(exception.value)

    def test_init_output_col_with_valid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        column_name = "input_column_name"
        # act
        for column_name in ["foo", "foo bar", "foo_bar", "   _"]:
            transformer = JSONTransformer(inputCol=column_name, 
                                          outputCol='output_column',
                                          removeInputCol=True,
                                          parser=mocked_service)

    def test_init_output_col_with_invalid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        column_name = None
        # act
        for column_name in [None, "", " ", "    "]:
            with pytest.raises(ValueError) as exception:
                JSONTransformer(inputCol='input_column', 
                                outputCol=column_name,
                                removeInputCol=True,
                                parser=mocked_service)
            # assert
            assert "output column name" in str(exception.value)
    
    def test_init_remove_input_column_with_valid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        valid_values = [0, 1, True, False]
        expect_values = [False, True, True, False]
        # act
        for i in range(len(valid_values)):
            transformer = JSONTransformer(inputCol='input_column', 
                                          outputCol='output_column',
                                          removeInputCol=valid_values[i],
                                          parser=mocked_service)
            # assert
            assert transformer.getRemoveInputCol() == expect_values[i]

    def test_init_remove_input_column_with_default_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        # act
        transformer =  JSONTransformer(inputCol='input column', 
                                        outputCol='output column',
                                        parser=mocked_service)
        # assert
        assert transformer.getRemoveInputCol() == False
    
    def test_init_valid_service(self):
        # arrange
        mocked_service = lambda x: x+1
        # act
        transformer =  JSONTransformer(inputCol='input column', 
                                        outputCol='output column',
                                        parser=mocked_service)
        # assert
        provided_service = transformer.getParser() 
        assert provided_service(10) == 11


    def test_init_none_callable_service(self):
        # arrange
        invalid_services = [None, 12, "12"]
        # act
        for i in range(len(invalid_services)):
            with pytest.raises(ValueError) as exinfo:
                JSONTransformer(inputCol='input column', 
                                outputCol='output column',
                                parser=invalid_services[i])
            # assert
            assert "parser instance" in str(exinfo.value) and "callable" in str(exinfo.value)


import types
from botocore.client import Config
import ibm_boto3
def __iter__(self): return 0

"""
"
" IBM COS reader
" boto3 is not thread safe(https://github.com/boto/botocore/issues/1246)
" create new session in each thread(ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html)
"
"""
def ibm_cos_reader(audio_file, bucket, token, endpoint):
    """
    @param::audio_file: the audio file uid
    @param::bukcet: the bucket name in which the audio file is stored
    @param::token: the API access token for IBM COS service
    @param::endpoint: the URL to access IBM COS service
    @return: the audio stream
    """
    session = ibm_boto3.session.Session() 
    cos_client = session.client(service_name='s3',
                                  ibm_api_key_id=token,
                                  ibm_auth_endpoint="https://iam.ng.bluemix.net/oidc/token",
                                  config=Config(signature_version='oauth'),
                                  endpoint_url=endpoint)
    audio_stream = cos_client.get_object(Bucket=bucket, Key=audio_file)['Body']
    if not hasattr(audio_stream, "__iter__"): audio_stream.__iter__ = types.MethodType( __iter__, audio_stream)
    return audio_stream


"""
"
" Local file reader
"
"""
def local_fs_reader(audio_file):
    """
    @param::audio_file: the full path including filename to the audio file
    @return: the filestream of audio file
    """
    return open(audio_file, "rb")


"""
"
" explode the struct column to multiple simple columns
" @ref: https://stackoverflow.com/questions/47669895/how-to-add-multiple-columns-using-udf?rq=1
"
"""

from pyspark import keyword_only
from pyspark.sql import functions as F
from pyspark.sql.types import StringType, FloatType, StructType, StructField, Row
from pyspark.sql import DataFrame
from pyspark.ml.pipeline import Transformer
from pyspark.ml.param.shared import HasInputCol, HasOutputCol, Param
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable



class FlatColumnTransformer(Transformer, 
                            HasInputCol, 
                            DefaultParamsReadable,
                            DefaultParamsWritable):
    
    """
    "
    " set init transformer and set parameters
    "
    """
    @keyword_only
    def __init__(self, 
                 inputCol=None):
        """
        @param::inputCol: the input column name contains sound file name
        @return: none
        """
        super(FlatColumnTransformer, self).__init__()
        kwargs = self._input_kwargs
        self._set(**kwargs)
    
    """
    "
    " set parameters, called from self._set(), inherited method.
    "
    """
    @keyword_only
    def setParams(self, inputCol=None):
        """
        @param::inputCol: the input column name contains sound file name
        @return: none
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    """
    "
    " set input columns name
    "
    """
    def setInputCol(self, value):
        """
        @param::value: the input columns name contains the sound file name
        @return: None
        """
        return self._set(inputCol=value)

    
    """
    "
    " perform the transform using provided IBM service api
    "
    """
    def _transform(self, df:DataFrame) -> DataFrame:
        """
        @param::df: the pyspark dataframe
        @return: the transformed dataframe
        """
        cols = df.columns
        if self.getInputCol() in cols:
            cols.remove(self.getInputCol())
            return df.withColumn('__explode_col_output__',F.explode(F.array(F.col(self.getInputCol())))) \
                     .select(*cols, "__explode_col_output__.*")
        else:
            raise ValueError("> FlatColumnTransformer class: inputCol is not in the dataframe.")



import json
from watson_transformer.contrib.response_base import ResponseBase
from pyspark.sql.types import StringType, FloatType, StructType, StructField, Row
from pyspark.sql import functions as F

"""
"
" default STT output interpreter which assume one alternative transcript and no speaker detection
"
"""

class DefaultSTTParser(ResponseBase):

    def __init__(self):
        super(DefaultSTTParser, self).__init__()


    """
    "
    " pass the json response to datatype
    "
    """
    def __call__(self, json_response):
        """
        @param::output: the output json object from STT
        @return:the transcript join by period in string format
        """
        if json_response:
            try:
                response = json.loads(json_response)
                if 'results'in response:
                    transcripts = [doc['alternatives'][0]['transcript'].strip() for doc in response['results']]
                    return '. '.join(transcripts) + '.'
                else:
                    # if result attribute is not present in the response 
                    # it is probably the speech recording has more than 30s in slient from the begining
                    return None
            except:
                return None
        else:
            return None

    """
    "
    " return the default STT return type
    "
    """
    def get_return_type(self):
        """
        @return: the defined return type
        """
        return  StringType()


"""
"
" pyspark transformer covert JSON data to data column(s)
" @ref: https://stackoverflow.com/questions/41399399/serialize-a-custom-transformer-using-python-to-be-used-within-a-pyspark-ml-pipel/52467470#52467470
" @ref: https://stackoverflow.com/questions/32331848/create-a-custom-transformer-in-pyspark-ml
"
"""
import pandas as pd
from concurrent.futures import ThreadPoolExecutor
from pyspark import keyword_only
from pyspark.sql import functions as F
from pyspark.sql.types import StringType, FloatType, StructType, StructField, Row
from pyspark.sql import DataFrame
from pyspark.ml.pipeline import Transformer
from pyspark.ml.param.shared import HasInputCol, HasOutputCol, Param
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable


class JSONTransformer(Transformer, 
                      HasInputCol, 
                      HasOutputCol,
                      DefaultParamsReadable,
                      DefaultParamsWritable):
    
    """
    "
    " set init transformer and set parameters
    "
    """
    @keyword_only
    def __init__(self, 
                 inputCol=None, 
                 outputCol=None, 
                 removeInputCol=False,
                 parser=None):
        """
        @param::inputCol: the input column name contains sound file name
        @param::outputCol: the output column name
        @param::removeInputCol: flag indicate whether remove input columns
        @param::parser: parser object parse JSON data to data column(s)
        @return: none
        """

        super(JSONTransformer, self).__init__()
        self.parser = Param(self, "parser", None)
        self._setDefault(parser=None)
        self.removeInputCol = Param(self, "removeInputCol", False)
        self._setDefault(removeInputCol=False)
        kwargs = self._input_kwargs
        self._set(**kwargs)

        # make sure parameter: token, endpoint set properly.
        if not callable(parser):
            raise ValueError('> The parser instance provided must be callable object.')
        if not inputCol or not inputCol.strip():
            raise ValueError('> The input column name is required.')
        if not outputCol or not outputCol.strip():
            raise ValueError('> The output column name is required.')



        # make sure parameter: token, endpoint set properly.
        if parser == None:
            raise ValueError('> JSONTransformer Class: JSON data parser object must be provided.')
    
    """
    "
    " set parameters, called from self._set(), inherited method.
    "
    """
    @keyword_only
    def setParams(self):
        """
        @param:: None
        @return: none
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    """
    "
    " set whether or not remove input column from dataframe
    "
    """
    def setRemoveInputCol(self, value):
        """
        @param::value: boolean value indcating enable vectorized udf 
        @return: None
        """
        return self._set(removeInputCol=value)
    
    """
    "
    " get enable state of vectorized udf
    "
    """
    def getRemoveInputCol(self):
        """
        @param:: None
        @return: vectorization flag
        """
        return self.getOrDefault(self.removeInputCol)
    
    """
    "
    " set the parser object
    "
    """
    def setParser(self, value):
        """
        @param::value: the IBM service API object
        @return: None
        """
        return self._set(parser=value)
    
    """
    "
    " get the parser object
    "
    """
    def getParser(self):
        """
        @param:None
        @return: the configured service object
        """
        return self.getOrDefault(self.parser)

    """
    "
    " set input columns name
    "
    """
    def setInputCol(self, value):
        """
        @param::value: the input columns name contains the sound file name
        @return: None
        """
        return self._set(inputCol=value)

    """
    "
    " set output columns name
    "
    """
    def setOutputCol(self, value):
        """
        @param::value: the output column name contains the output result
        @return: None
        """
        return self._set(outputCol=value)

    
    """
    "
    " perform the transform 
    "
    """
    def _transform(self, df:DataFrame) -> DataFrame:
        """
        @param::df: the pyspark dataframe
        @return: the transformed dataframe
        """
        # get the new service instance
        parser = self.getParser()
        outputCol = self.getOutputCol()
        inputCol = self.getInputCol()
        removeInputCol = self.getRemoveInputCol()
        return_type = parser.get_return_type()
        # CPU bounded task, not going to benefit from vectorized UDF very much
        parser_udf = F.udf(parser, return_type).asNondeterministic()
        df = df.withColumn(outputCol, parser_udf(F.col(inputCol)))
        # drop input column based on flag
        if removeInputCol:
            df = df.drop(self.getInputCol())
        return df


"""
"
" IBM STT Service Executable Class
" docu: https://cloud.ibm.com/apidocs/speech-to-text
"
"""

import json
from pyspark import keyword_only
from ibm_watson import SpeechToTextV1
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_cloud_sdk_core.api_exception import ApiException
from pyspark.sql.types import StringType
from watson_transformer.service.service_base import ServiceBase

class STT(ServiceBase):
    
    @keyword_only
    def __init__(self, token, endpoint, reader, strict_mode=True, **params):
        """
        @param::token: the IBM STT API access token
        @param::endpoint: the endpoint url for the STT API
        @param::reader: the object read audio stream using audio file name/id
        @param::params: the kv params passing to underlying SpeechToTextV1 constructor
        @return: the output formatted by formatter executable
        """
        super(STT, self).__init__(strict_mode)
        self.token = token
        self.endpoint = endpoint
        self.reader = reader
        self.params = params

    def __call__(self, audio_file):
        """
        @param::audio_file: the audio filename/id for reader to retrieve the audio stream
        @return: the output formatted by formatter object
        """
        if audio_file:
            # load asset
            audio_stream = self.reader(audio_file)
            # check if audio stream is valid
            if not audio_stream:
                return None

            # init stt client
            authenticator = IAMAuthenticator(self.token)
            stt = SpeechToTextV1(authenticator=authenticator)
            stt.set_service_url(self.endpoint)
            
            # send the request
            try:
                response = stt.recognize(audio=audio_stream,**self.params).get_result()
            except ApiException as api_ex:
                response = {'api_error_message': str(api_ex)} # less likely recoverable if it is STT API error
            except Exception as ex:
                if self.strict_mode:
                    raise RuntimeError("*** runtime error caused by input: '%s'"%(audio_file)) # maybe recoverable by retry
                else:
                    response = {'error_message': str(ex)}
            return json.dumps(response) if response else None
        else:
            return None
    
    def get_return_type(self):
        """
        @param::output_col: output column name
        @return: the output type struct
        """
        return StringType()

    def get_new_client(self):
        return STT(token = self.token, 
                   endpoint = self.endpoint,
                   reader = self.reader,
                   strict_mode = self.strict_mode,
                   **self.params)


"""
"
" define the contract/interfact 
"
"""

class ServiceBase():
    def __init__(self, strict_mode=True):
        self.strict_mode = strict_mode

    def __call__(self, data):
        raise NotImplementedError('> service class __call__ method is not implemented.')

    def get_return_type(self, data):
        raise NotImplementedError('> service class get_return_type method is not implemented.')

    def get_new_client(self):
        raise NotImplementedError('> service class get_new_client method is not implemented.')




"""
"
" test watson service transformer class
"
"""
import pytest
from unittest import mock
from watson_transformer.watson_service_transformer import WatsonServiceTransformer



@pytest.fixture(scope='function')
def mock_service(request):
    # mock the service
    service = mock.MagicMock(return_value='foo response')
    return lambda : service


class TestWatsonServiceTransformer():

    def test_init_input_col_with__valid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        column_name = "input_column_name"
        # act
        for column_name in ["foo", "foo bar", "foo_bar", "   _"]:
            transformer = WatsonServiceTransformer(inputCol=column_name, 
                                                   outputCol='output_column',
                                                   vectorization=True,
                                                   max_workers = 10,
                                                   service=mocked_service)
            # assert
            assert transformer.getInputCol() == column_name

    def test_init_input_col_with_invalid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        column_name = None
        # act
        for column_name in [None, "", " ", "    "]:
            with pytest.raises(ValueError) as exception:
                _ = WatsonServiceTransformer(inputCol=column_name, 
                                            outputCol='output_column',
                                            vectorization=True,
                                            max_workers = 10,
                                            service=mocked_service)
            # assert
            assert "input column name" in str(exception.value)

    def test_init_output_col_with_valid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        column_name = "input_column_name"
        # act
        for column_name in ["foo", "foo bar", "foo_bar", "   _"]:
            transformer =  WatsonServiceTransformer(inputCol='input column', 
                                                    outputCol=column_name,
                                                    vectorization=True,
                                                    max_workers = 10,
                                                    service=mocked_service)
            # assert
            assert transformer.getOutputCol() == column_name

    def test_init_output_col_with_invalid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        column_name = None
        # act
        for column_name in [None, "", " ", "    "]:
            with pytest.raises(ValueError) as exception:
                WatsonServiceTransformer(inputCol="input column", 
                                         outputCol=column_name,
                                         vectorization=True,
                                         max_workers = 10,
                                         service=mocked_service)
            # assert
            assert "output column name" in str(exception.value)

    def test_init_max_workers_with_valid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        max_workers = [1,2,3,4,5]
        expect_max_workers = [1,2,3,4,5]
        # act
        for i in range(len(max_workers)):
            transformer =  WatsonServiceTransformer(inputCol='input column', 
                                                    outputCol='output column',
                                                    vectorization=True,
                                                    max_workers = max_workers[i],
                                                    service=mocked_service)
            # assert
            assert transformer.getMax_workers() == expect_max_workers[i]

    def test_init_max_workers_with_default_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        expected_default_value = 5
        # act
        transformer =  WatsonServiceTransformer(inputCol='input column', 
                                                outputCol='output column',
                                                vectorization=True,
                                                service=mocked_service)
        # assert
        assert transformer.getMax_workers() == expected_default_value

    def test_init_max_workers_with_invalid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        max_workers = [None, "1", -1, 0, -100]
        # act
        for i in range(len(max_workers)):
            with pytest.raises(ValueError) as exinfo:
                WatsonServiceTransformer(inputCol='input column', 
                                         outputCol='output column',
                                         vectorization=True,
                                         max_workers = max_workers[i],
                                         service=mocked_service)
            # assert
            assert "maximum workers" in str(exinfo.value)
    
    def test_init_vectorization_with_valid_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        vectorizations = [0, 1, True, False]
        expect_vectorizations = [False, True, True, False]
        # act
        for i in range(len(vectorizations)):
            transformer =  WatsonServiceTransformer(inputCol='input column', 
                                                    outputCol='output column',
                                                    vectorization=vectorizations[i],
                                                    max_workers = 10,
                                                    service=mocked_service)
            # assert
            assert transformer.getVectorization() == expect_vectorizations[i]

    def test_init_vectorization_with_default_value(self, mock_service):
        # arrange
        mocked_service = mock_service()
        # act
        transformer =  WatsonServiceTransformer(inputCol='input column', 
                                                outputCol='output column',
                                                service=mocked_service)
        # assert
        assert transformer.getVectorization() == False
    
    def test_init_valid_service(self):
        # arrange
        mocked_service = lambda x: x+1
        # act
        transformer =  WatsonServiceTransformer(inputCol='input column', 
                                                outputCol='output column',
                                                service=mocked_service)
        # assert
        provided_service = transformer.getService() 
        assert provided_service(10) == 11


    def test_init_none_callable_service(self):
        # arrange
        values = [None, 12, "12"]
        # act
        for i in range(len(values)):
            with pytest.raises(ValueError) as exinfo:
                WatsonServiceTransformer(inputCol='input column', 
                                         outputCol='output column',
                                         service=values[i])
            # assert
            assert "service instance" in str(exinfo.value) and "callable" in str(exinfo.value)


from watson_transformer.watson_service_transformer import WatsonServiceTransformer
from watson_transformer.flat_column_transformer import FlatColumnTransformer
from watson_transformer.json_transformer import JSONTransformer


"""
"
" response base class
"
"""
class ResponseBase():

    def __call__(self, json_data):
        raise NotImplementedError('> ResponseBase class: __call__ method is not implemented.')

    def get_return_type(self):
        raise NotImplementedError('> ResponseBase class: get_return_type method is not implemented.')
